#!/usr/bin/perl -w

#####################################################################
# saves3.pl
#
# This is a simple utility you can call to backup an object bucket
# (e.g., S3) that's been mounted using s3fs-fuse, via NetWorker.
# There's nothing "S3-specific" or even "object specific" about it.
# When NetWorker is traversing a FUSE mounted filesystem for object
# storage, it seems to detect an inode change on directories and
# doesn't traverse into them.
#
# To avoid this, we can use the '-x' option, which tells NetWorker
# to follow mount-points. NetWorker will still report it sees an
# inode change, but it will walk into the directory.
#
# This script allows you to do simulated full and incremental backups;
# while a 'save -x' must be a manual backup, and normally a manual
# backup is "everything on the system", we can still give a save -x
# a timestamp of a previous backup. When you do this, you'll get a
# pseudo-incremental. (This will let you get a filesystem point of
# time view, during recoveries, too.)
#
# So the real intent of the script is to allow you to run full and
# 'incr' manual backups.
#
# Notes:
# 1. Run "saves3.pl -h" for full usage information.
# 2. No tests have been done against mount points with spaces.
# 3. PSS isn't used in this.
#
# THIS SCRIPT IS SUPPLIED WITH NO WARRANTY. YOU ARE RESPONSIBLE FOR
# ANY DATA YOU PROTECT WITH THIS SCRIPT AND YOU SHOULD VERIFY ITS
# SUCCESSFUL OPERATION THOROUGHLY BEFORE USING FULLY. YOU ARE
# RESPONSIBLE FOR ANY RUN OR EGRESS FEES IN, OUT OF OR BETWEEN
# CLOUD SERVICES.
#
# Preston de Guise, Data Protection Blog, https://nsrd.info/blog
#####################################################################

######################
# Modules
######################
use strict;
use Getopt::Std;
use File::Basename;
use Sys::Hostname;


######################
# Global Variables
######################
my $server = hostname();
my $client = $server;
my $self = basename($0);
my $pool = "Default";
my $level = "full";
my $path = "none";
my $version = "1";
my $prevTimeStamp = "";
my $logDir = "null";
my $expDate = "+31 days";
my $logging = 0;
my $simpleSelf = $self;
if ($simpleSelf =~ /^([^\.]*)\..*$/) {
	$simpleSelf = $1;
}
my ($year,$month,$day,$hour,$minute,$second) = (localtime(time))[5,4,3,2,1,0];
$year += 1900;
$month++;
$month = sprintf("%02d",$month);
$day = sprintf("%02d",$day);
$hour = sprintf("%02d",$hour);
$minute = sprintf("%02d",$minute);
$second = sprintf("%02d",$second);
my $logFileName = "/nsr/applogs/${simpleSelf}_$year-$month-$day-$hour$minute$second.log";


######################
# Subroutines
######################

# usage([@messages]) prints usage information, any additional messages,
# and then exits.
sub usage {
	print <<EOF;

Usage: $self [-h] [-v] -d dir -b pool -s server [-l lvl] [-D] [-c client] [-e exp]

Where:

-h        Print this help and exit.
-v        Print version number and exit.
-d dir    Path that S3 bucket is mounted at.
-b pool   NetWorker pool to write backup to.
-s server NetWorker server to talk to (defaults to this host)
-c client NetWorker client to associate the backup with (defaults to this host)
-l lvl    Backup level (full, or incr - defaults to full)
-e exp    Expiration date in valid nsr_date format. Defaults to \"$expDate\"
-D        Go into debug logging mode. (Writes to /nsr/applog).

Performs a backup of an object store (e.g,. AWS S3) mounted via
s3fs-fuse to the nominated NetWorker server/pool. While technically
a manual backup, if 'incr' is specified as a level, we look for
the most recent backup of the same path for this client and
tag to that backup's savetime.

This script requires the NetWorker extended client package to be
installed. See comments at start of script for additional information.
EOF

	if (@_+0 != 0) {
		my @messages = @_;
		foreach my $message (@messages) {
			my $tmp = $message;
			chomp $tmp;
			print "$tmp\n";
		}
	}
	die "\n";
}

# print_or_log($message) prints the message if logging is off, or
# writes it to the log file if logging is on.
sub print_or_log {
	my $message = $_[0];
	my $tmp = $message;
	chomp $tmp;

	if ($logging) {
		if (open(LOG,">>$logFileName")) {
			print LOG "$tmp\n";
			close(LOG);
		} else {
			print ("(Logging failed, sending to stdout)\n$tmp\n");
		}
	} else {
		print ("$tmp\n");
	}
}

# get_last_timestamp(client,server,pool,path) gets the last timestamp for
# the specified path from NetWorker server 'server' for client 'cleint' in
# pool 'pool'.
sub get_last_timestamp {
	if (@_+0 != 4) {
		my $argCount = @_+0;
		die "get_last_timestamp\nSomething really odd happened. Expected 4 arguments, got $argCount\n";
	} 

	# else...
	my $nsrClient = $_[0];
	my $nsrServer = $_[1];
	my $nsrPool = $_[2];
	my $dirPath = $_[3];

	my $mminfo = "mminfo -s $nsrServer -q \"client=$nsrClient,pool=$nsrPool,name=$dirPath\" -r nsavetime -otR";
	my $nsavetime = "undef";
	if (open(MMI,"$mminfo 2>&1 |")) {
		while (<MMI>) {
			my $line = $_;
			chomp $line;
			if ($line =~ /^\s*(\d+)$/) {
				# Bingo, we've found the most recent backup for this.
				# The recent backup will always be the timestamp to 
				# chain onto for an incremental backup.
				$nsavetime = $1;
				last;
			}
		}
		close(MMI);
	} else {
		die ("An error occurred while running:\n\n$mminfo\n\n");
	}

	return $nsavetime;
}

# do_backup(client,server,pool,path,level,prevTimeStamp,expDate) runs the
# requested backup and reports back.
sub do_backup {
	if (@_+0 != 7) {
		my $argCount = @_+0;
		die "do_backup\nSomething really odd happened. Exected 7 arguments, got $argCount\n";
	}

	# else...
	my $nsrClient = $_[0];
	my $nsrServer = $_[1];
	my $nsrPool = $_[2];
	my $backupPath = $_[3];
	my $nsrLevel = $_[4];
	my $prevTS = $_[5];
	my $nsrExpire = $_[6];

	my $tsInsert = "";

	if ($prevTS =~ /^\d+$/) {
		$tsInsert = " -t $prevTS ";
	}

	my $saveCommand = "save -x $tsInsert -s $nsrServer -c $nsrClient -b $nsrPool -e \"$expDate\" $backupPath";

	if ($logging) {
		print_or_log("Calling backup\n\n$saveCommand\n\n\n");
	}

	# Add logging here.
	my $retVal = 0;
	if ($logging) {
		$retVal = system("$saveCommand >> $logFileName 2>&1");
	} else {
		$retVal = system("$saveCommand > /dev/null 2>&1");
	}
	if ($retVal == 0) {
		return 1;
	} else {
		return 0;
	}
}


# Main
my %opts;
if (getopts('hvc:s:l:b:d:e:D',\%opts)) {
	$logging = 1 if (defined($opts{D}));
	usage() if (defined($opts{h}));
	die "$self is v$version\n" if (defined($opts{v}));

	$server = $opts{s} if (defined($opts{s}));
	$client = $opts{c} if (defined($opts{c}));
	if (defined($opts{l})) {
		my $interimLevel = lc($opts{l});
		if ($interimLevel eq "full" || $interimLevel eq "incr") {
			$level = $interimLevel;
		} else {
			usage("Level $level supplied - not a recognised option.");
		}
	}

	if (defined($opts{d})) {
		if (-d $opts{d}) {
			$path = $opts{d};
		} else {
			usage("Specified path does not exist: $opts{d}\n");
		}
	} else {
		usage("You must specify a mount point to backup via -d path");
	}

	if (defined($opts{b})) {
		$pool = $opts{b};
	} else {
		print_or_log("NOTE: No pool specified, using 'Default'. Are you sure you meant that?");
		sleep(5);
	}

	if (defined($opts{e})) {
		$expDate = $opts{e};
	} else {
		print_or_log("NOTE: No expiration date specified. Defaulting to $expDate");
	}
}

if ($level eq "incr") {
	$prevTimeStamp = get_last_timestamp($client,$server,$pool,$path);
	print_or_log ("NOTE: Last timestamp for $server:$client:$pool:$path = $prevTimeStamp");
}

if (do_backup($client,$server,$pool,$path,$level,$prevTimeStamp,$expDate)) {
	print_or_log ("Backup successful");
	exit 0;
} else {
	die "Failed to run the requested backup. Check logs for more details.\n";
}
